;; Generated package description from jumpc.el  -*- no-byte-compile: t -*-
(define-package "jumpc" "3.0.0.20221212.224814" "jump to previous insertion points" 'nil :commit "8ce5956123f50b389c602eb83eb190138bb5f2d6" :url "https://elpa.gnu.org/packages/jumpc.html" :authors '(("Ivan Kanis" . "ivan@kanis.fr")) :maintainer '("Ivan Kanis" . "ivan@kanis.fr"))
